@@ -4,11 +4,10 @@
 
 ## ------------------------------------------------------------------------
 library(ProFit)
-library(knitr)
 
 ## ------------------------------------------------------------------------
 data('ExampleInit', package="ProFit")
-kable(head(ExampleInit, 10))
+head(ExampleInit, 10)
 
 ## ------------------------------------------------------------------------
 datasource='KiDS'
