@@ -32,24 +32,42 @@
 
 namespace profit {
 
-invalid_parameter::invalid_parameter(const std::string &what_arg) :
-	exception(),
+exception::exception(const std::string &what_arg) :
 	m_what(what_arg)
 {
 	// no-op
 }
 
+exception::~exception() throw () {
+	// no-op
+}
+
+const char *exception::what() const throw() {
+	return m_what.c_str();
+}
+
+invalid_parameter::invalid_parameter(const std::string &what_arg) :
+	exception(what_arg)
+{
+	// no-op
+}
+
 invalid_parameter::~invalid_parameter() throw () {
 	// no-op
 }
 
-const char *invalid_parameter::what() const throw() {
-	return m_what.c_str();
+unknown_parameter::unknown_parameter(const std::string &what_arg) :
+	invalid_parameter(what_arg)
+{
+	// no-op
+}
+
+unknown_parameter::~unknown_parameter() throw () {
+	// no-op
 }
 
 opencl_error::opencl_error(const std::string &what_arg) :
-	exception(),
-	m_what(what_arg)
+	exception(what_arg)
 {
 	// no-op
 }
@@ -58,14 +76,8 @@ opencl_error::~opencl_error() throw () {
 	// no-op
 }
 
-const char *opencl_error::what() const throw() {
-	return m_what.c_str();
-}
-
-
 fft_error::fft_error(const std::string &what_arg) :
-	exception(),
-	m_what(what_arg)
+	exception(what_arg)
 {
 	// no-op
 }
@@ -74,8 +86,13 @@ fft_error::~fft_error() throw () {
 	// no-op
 }
 
-const char *fft_error::what() const throw() {
-	return m_what.c_str();
+fs_error::fs_error(const std::string &what_arg) :
+	exception(what_arg)
+{
+	// no-op
 }
 
+fs_error::~fs_error() throw () {
+	// no-op
+}
 } /* namespace profit */
